package com.github.polpetta.mezzotre.telegram.command;

import com.github.polpetta.mezzotre.i18n.TemplateContentGenerator;
import com.github.polpetta.mezzotre.orm.model.TgChat;
import com.github.polpetta.mezzotre.orm.telegram.CampaignUtil;
import com.github.polpetta.mezzotre.orm.telegram.ChatUtil;
import com.github.polpetta.mezzotre.util.UUIDGenerator;
import com.pengrad.telegrambot.model.Update;
import com.pengrad.telegrambot.model.request.ParseMode;
import com.pengrad.telegrambot.request.BaseRequest;
import com.pengrad.telegrambot.request.SendMessage;
import java.util.Collections;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Executor;
import javax.inject.Inject;
import javax.inject.Named;

// FIXME tests, doc
public class CreateCampaign implements Processor {

  private static final String TRIGGERING_KEYWORD = "/createCampaign";
  // FIXME create the label
  private static final String BUTTON_CREATE_IT_LOCALE_KEY = "createCampaign.createItButton";
  private static final String CAMPAIGN_NAME_CHAT_CTX_FIELD = "campaign_name";
  private static final String CAMPAIGN_NAME_TEMPLATE_CTX_FIELD = "campaignName";
  private static final int MAX_CAMPAIGN_NAME_LENGTH = 256;
  private static final int MAX_CAMPAIGN_DESCRIPTION_LENGTH = 4096;
  private final TemplateContentGenerator templateContentGenerator;
  private final Executor threadPool;
  private final UUIDGenerator uuidGenerator;
  private final ChatUtil chatUtil;
  private final CampaignUtil campaignUtil;

  @Inject
  public CreateCampaign(
      TemplateContentGenerator templateContentGenerator,
      @Named("eventThreadPool") Executor threadPool,
      UUIDGenerator uuidGenerator,
      ChatUtil chatUtil,
      CampaignUtil campaignUtil) {

    this.templateContentGenerator = templateContentGenerator;
    this.threadPool = threadPool;
    this.uuidGenerator = uuidGenerator;
    this.chatUtil = chatUtil;
    this.campaignUtil = campaignUtil;
  }

  @Override
  public Set<String> getTriggerKeywords() {
    return Set.of(TRIGGERING_KEYWORD, "/newCampaign");
  }

  @Override
  public String getLocaleDescriptionKeyword() {
    return "createCampaign.cmdDescription";
  }

  @Override
  public CompletableFuture<Optional<BaseRequest<?, ?>>> process(TgChat chat, Update update) {
    // There are multiple steps, we have to route the incoming campaign creation to the right one
    return switch ((int) chat.getChatContext().getStep()) {
      default -> promptCampaignName(chat);
      case 1 -> verifyCampaignNameAndPromptCampaignDescription(chat, update);
      case 2 -> verifyDescriptionAndCompleteCreation(chat, update);
    };
  }

  private CompletableFuture<Optional<BaseRequest<?, ?>>> promptCampaignName(TgChat chat) {
    return CompletableFuture.supplyAsync(
            () ->
                templateContentGenerator.mergeTemplate(
                    ctx -> {}, chat.getLocale(), "template/telegram/createCampaign.0.vm"),
            threadPool)
        .thenApplyAsync(
            message -> {
              chatUtil.updateChatContext(chat, TRIGGERING_KEYWORD, 1, Collections.emptyMap());

              return Optional.of(
                  new SendMessage(chat.getId(), message).parseMode(ParseMode.Markdown));
            },
            threadPool);
  }

  private CompletableFuture<Optional<BaseRequest<?, ?>>>
      verifyCampaignNameAndPromptCampaignDescription(TgChat chat, Update update) {
    /*
    Flow:
    1 - Input validation (name length, etc...)
    2 - Build message according to validation output
    3 - Send message back
    4 - ???
    5 - Profit
    */
    return CompletableFuture.completedFuture(chat)
        .thenApply(
            // 1 - Perform input validation
            ignored ->
                Util.extractText(update)
                    .filter(campaignName -> campaignName.length() <= MAX_CAMPAIGN_NAME_LENGTH))
        .thenApplyAsync(
            nameOpt -> {
              // 2 - create appropriate message and update chat context
              return nameOpt
                  .map(
                      campaignName -> {
                        chatUtil.updateChatContext(
                            chat,
                            TRIGGERING_KEYWORD,
                            2,
                            Collections.singletonMap(CAMPAIGN_NAME_CHAT_CTX_FIELD, campaignName));
                        return templateContentGenerator.mergeTemplate(
                            ctx -> ctx.put(CAMPAIGN_NAME_TEMPLATE_CTX_FIELD, campaignName),
                            chat.getLocale(),
                            "template/telegram/createCampaign.1.vm");
                      })
                  // We don't update the context here on purpose. If the user wants to try another
                  // campaign name is free to do so!
                  .orElseGet(
                      () ->
                          templateContentGenerator.mergeTemplate(
                              ctx -> {},
                              chat.getLocale(),
                              "template/telegram/createCampaign.nameNotValid.vm"));
            })
        .thenApply(
            // 3 - send message back
            message ->
                Optional.of(new SendMessage(chat.getId(), message).parseMode(ParseMode.Markdown)));
  }

  private CompletableFuture<Optional<BaseRequest<?, ?>>> verifyDescriptionAndCompleteCreation(
      TgChat chat, Update update) {
    /*
    Flow:
    1 - Input validation (description length, etc...)
    2 - Build message according to validation output
    3 - Send message back
    4 - ???
    5 - Profit
    */
    return CompletableFuture.completedFuture(chat)
        .thenApply(
            ignored ->
                Util.extractText(update)
                    .filter(description -> description.length() <= MAX_CAMPAIGN_DESCRIPTION_LENGTH))
        .thenApplyAsync(
            descriptionOpt ->
                descriptionOpt
                    .map(
                        campaignDescription -> {
                          // We go full beans here because we know that the property will never be
                          // null at this point - if it is null, then there are bigger problems, and
                          // we should start questioning our life (and our programming skills
                          // really)
                          final String campaignName =
                              (String)
                                  chat.getChatContext()
                                      .getAdditionalProperties()
                                      .get(CAMPAIGN_NAME_CHAT_CTX_FIELD);
                          campaignUtil.insertNewCampaign(campaignName, campaignDescription);
                          chatUtil.updateChatContext(
                              chat, TRIGGERING_KEYWORD, 3, Collections.emptyMap());
                          return templateContentGenerator.mergeTemplate(
                              ctx -> {}, chat.getLocale(), "template/telegram/createCampaign.2.vm");
                        })
                    .orElseGet(
                        () ->
                            templateContentGenerator.mergeTemplate(
                                ctx -> {},
                                chat.getLocale(),
                                "template/telegram/createCampaign.descriptionNotValid.vm")))
        .thenApply(
            message ->
                Optional.of(new SendMessage(chat.getId(), message).parseMode(ParseMode.Markdown)));
  }
}
